IMPLEMENTATION MODULE RealConversions;

(**************************************************************************
   JAVA      Modula-2 utility for float conversions to/from strings.

   Copyright (c) 1998 by Mill Hill & Canterbury Corporation, Ltd.
**************************************************************************)

(*$XL+       Modula-2 language extensions: allow typed constants         *)

FROM   SYSTEM           IMPORT ADR, TSIZE;
IMPORT Strings;
IMPORT Conversions;

CONST
  SGLDIG           = 6;        (* max # of significand dec. digits   *)
  DBLDIG           = 15;       (* max # of significand dec. digits   *)
  DIGMAX           = DBLDIG*2; (* max # of digits in string          *)
                               (* ( *2 is a good fudge factor)       *)
  DIGPREC          = DBLDIG+2; (* max # of significant digits        *)
                               (* (+2 for fractional part after min  *)
                               (*     significant digits)            *)
  IsLongReal       = (TSIZE( REAL ) = TSIZE( LONGREAL ));


CONST
  negtab  : ARRAY OF LONGREAL =
  [1.E-256, 1.E-128, 1.E-64, 1.E-32, 1.E-16, 1.E-8, 1.E-4, 1.E-2, 1.E-1, 1.0];
  postab  : ARRAY OF LONGREAL =
  [1.E+256, 1.E+128, 1.E+64, 1.E+32, 1.E+16, 1.E+8, 1.E+4, 1.E+2, 1.E+1];

TYPE
  DIGITSTRING      = ARRAY [0..DIGMAX-1 +1 +1] OF CHAR;
                     (* DIGMAX digits in a string   *)
                     (* +1 for end of string        *)
                     (* +1 in case rounding adds    *)


PROCEDURE IsLongNAN( r : LONGREAL ) : BOOLEAN;
BEGIN
  SYSTEM.INLINE
  (
    return java.lang.Double.isNaN(r);
  );
END IsLongNAN;


PROCEDURE IsShortNAN( r : SHORTREAL ) : BOOLEAN;
BEGIN
  SYSTEM.INLINE
  (
    return java.lang.Float.isNaN(r);
  );
END IsShortNAN;


PROCEDURE IsLongINFINITY( r : LONGREAL ) : BOOLEAN;
BEGIN
  SYSTEM.INLINE
  (
    return java.lang.Double.isInfinite( r );
  );
END IsLongINFINITY;


PROCEDURE IsShortINFINITY( r : SHORTREAL ) : BOOLEAN;
BEGIN
  SYSTEM.INLINE
  (
    return java.lang.Float.isInfinite( r );
  );
END IsShortINFINITY;


PROCEDURE IsLongINDEFINITE( r : LONGREAL ) : BOOLEAN;
BEGIN
  SYSTEM.INLINE
  (
    return (java.lang.Double.doubleToLongBits( r ) == 0xFFF8000000000000L);
  );
END IsLongINDEFINITE;


PROCEDURE IsShortINDEFINITE( r : SHORTREAL ) : BOOLEAN;
VAR
  i : LONGINT;
BEGIN
  SYSTEM.INLINE
  (
    return (java.lang.Float.floatToIntBits( r ) == 0xFFC00000);
  );
END IsShortINDEFINITE;



PROCEDURE COPY( source : ARRAY OF CHAR; VAR dest : ARRAY OF CHAR );
BEGIN
  dest[0] := 0C;
  Strings.Append( source, dest );
END COPY;



PROCEDURE Fill
(
  Pos              : LONGINT;
  Size             : LONGINT;
  Filler           : CHAR;
  VAR Str          : ARRAY OF CHAR
);
VAR
  i                : LONGINT;
  j                : LONGINT;
  highStr          : LONGINT;
BEGIN
  highStr := HIGH(Str);
  i := Strings.Length( Str );
  IF Pos > i THEN
    RETURN;
  END;
  IF Pos = i THEN
    WHILE (i < highStr) & (Size > 0) DO
      Str[ i ] := Filler;
      INC( i );
      DEC( Size );
    END;
    Str[ i ] := 0C;
    RETURN;
  END;
  IF Pos + Size >= highStr THEN
    Size := highStr - Pos;
    WHILE Size > 0 DO
      Str[ Pos ] := Filler;
      INC( Pos );
      DEC( Size );
    END;
    Str[ Pos ] := 0C;
    RETURN;
  END;
  j := i - Pos;
  IF Pos + Size + j >= highStr THEN
    j := highStr - Pos - Size;
  END;
  i := Pos + Size + j;
  Str[ i ] := 0C;
  WHILE j > 0 DO
    DEC( j );
    DEC( i );
    Str[ i ] := Str[ i-Size ];
  END;
  i := Pos;
  WHILE i < Pos+Size DO
    Str[ i ] := Filler;
    INC( i );
  END;
END Fill;



PROCEDURE floatcvt
(
  isLong      : BOOLEAN;
  cnvflag     : BOOLEAN;
  r           : LONGREAL;
  digits      : INTEGER;
  VAR decpt   : INTEGER;
  VAR sign    : BOOLEAN;
  VAR result  : DIGITSTRING;
  VAR done    : BOOLEAN
);
VAR
  i           : INTEGER;
  pow         : INTEGER;
  digstr      : DIGITSTRING;
  p           : INTEGER;
  n           : LONGINT;
  digprec     : INTEGER;
  highResult  : LONGINT;
BEGIN
(*
   Convert a long real value to a string of decimal digits.
   Input:
     isLong    = up to (15 if TRUE; 6 if FALSE) digits are significand
     cnvflag = how to interpret wanted # of digits
     r       = long real value to be converted
     digits  = if cnvflag = FALSE:
                 # of digits in resulting string past the decimal point
               if cnvflag = TRUE:
                 # of digits in resulting string
   Returns:
     decpt  = position of decimal point from left of first digit
     sign   = TRUE if value was negative
     result = string with digits
   BUGS:
     This routine will hang if it is passed a NAN or INFINITY.
*)
  highResult := HIGH(result);
  IF r < 0.0 THEN
    sign := TRUE;
    r := -r;
  ELSE
    sign := FALSE;
  END;
  IF digits < 0 THEN
    digits := 0;
  ELSIF digits >= DIGMAX THEN
    digits := DIGMAX;
  END;
  IF r = 0.0 THEN
    result[ 0 ] := 0C;
    Fill( 0, highResult, '0', result );
    decpt := 0;
    done := TRUE;
    RETURN;
  END;

  (* Adjust things so that 1 <= r < 10 *)
  decpt := 1;
  pow := 256;
  i := 0;
  WHILE r < 1. DO
    WHILE r < negtab[ i+1 ] DO
      r := r * postab[ i ];
      decpt := decpt - pow;
    END;
    pow := pow DIV 2;
    i := i + 1;
  END;
  pow := 256;
  i := 0;
  WHILE r >= 10. DO
    WHILE r >= postab[ i ] DO
      r := r * negtab[ i ];
      decpt := decpt + pow;
    END;
    pow := pow DIV 2;
    i := i + 1;
  END;

  IF cnvflag THEN
    digits := digits + decpt;
    IF digits < 0 THEN
      digits := 0;
    ELSIF digits >= DIGMAX THEN
      digits := DIGMAX;
    END;
  END;

  (* Pick off digits 1 by 1 and stuff into digstr[] *)
  (* Do 1 extra digit for rounding purposes         *)
  IF isLong THEN
    digprec := DBLDIG+1;
  ELSE
    digprec := SGLDIG+1;
  END;
  FOR p := 0 TO digits BY 1 DO
    digstr[p] := 0C;
    IF p >= digprec THEN
      (* only 'digprec' digits are significand *)
      digstr[ p ] := '0';
    ELSE
      (* store digit before getting next one *)
      n := TRUNC( r );
      digstr[ p ] := CHR( n + ORD('0') );
      r := (r - LONGFLOAT(n)) * 10.0;
    END;
  END;

  IF digits < digprec THEN
    p := digits;
  ELSE
    p := digprec-1;
  (*p := digprec;*)
  END;
  IF digstr[ p ] >= '5' THEN
    (* we need to round *)
    LOOP
      IF p = 0 THEN (* if at start *)
        IF cnvflag THEN
          digits := digits + 1;
        END;
        decpt := decpt + 1; (* shift dec pnt *)
        digstr[ 0 ] := '1'; (* "100000..."   *)
        EXIT;
      END;
      digstr[ p ] := '0';
      p := p - 1;
      IF digstr[ p ] <> '9' THEN
        digstr[p] := CHR( ORD( digstr[p] )+1 );
        EXIT;
      END;
    END;
  END;

  digstr[ digits ] := 0C;  (* terminate string *)
  COPY( digstr, result );
  done := TRUE;
END floatcvt;



PROCEDURE FloatToFixPt
(
  isLong   : BOOLEAN;
  r        : LONGREAL;
  size     : INTEGER;
  digits   : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
VAR
  i       : INTEGER;
  p       : INTEGER;
  decpt   : INTEGER;
  sign    : BOOLEAN;
  buf     : DIGITSTRING;
  j       : INTEGER;
  ch      : CHAR;
BEGIN
  (*
     Convert a long real value into a fixed floating point format:
     ["-"] digit {digit} "." {digit}
  *)

  IF IsLongINFINITY(r) THEN
    COPY( "infinity", s );
    done := FALSE;
    RETURN;
  ELSIF IsLongNAN(r) THEN
    COPY( "NaN", s );
    done := FALSE;
    RETURN;
  END;

  (* make sure 'size' and 'digits' are in proper ranges *)
  IF r < 0. THEN
    i := 1;
  ELSE
    i := 0;
  END;
  IF VAL( CARDINAL, size ) > SYSTEM.LEN( s )-1 THEN
    (* size does not fit into string *)
    s[0] := 0C;
    done := FALSE;
    RETURN;
  END;
  IF size < i + 2 THEN
    (* at least ["-"] digit "." needed for result string *)
    s[0] := 0C;
    done := FALSE;
    RETURN;
  END;
  IF digits > DIGMAX THEN
    digits := DIGMAX;
  ELSIF digits < 0 THEN
    done := FALSE;
    RETURN;
  END;
  IF size - i - 2 < digits THEN
    (* too many digits wanted after dec. point *)
    s[0] := 0C;
    done := FALSE;
    RETURN;
  END;
  s[ size ] := 0C;

  (* store significand digits into a buffer *)
  floatcvt( isLong, TRUE, r, digits, decpt, sign, buf, done );

  p := 0;
  i := 0;
  IF i >= size THEN
    done := FALSE;
    RETURN;
  END;
  IF sign THEN
    s[i] := '-';
    INC( i );
  END;
  IF i >= size THEN
    done := FALSE;
    RETURN;
  END;

  IF decpt <= 0 THEN
    (* at least 1 digit before dec point *)
    s[i] := '0';
    INC( i );
  END;
  IF i >= size THEN
    done := FALSE;
    RETURN;
  END;

  (* store all digits before dec. point into result string *)
  ch := buf[ p ];
  WHILE (decpt > 0) AND (i < size) AND (ch <> 0C) DO
    s[i] := buf[p];
    INC( i );
    INC( p );
    DEC( decpt );
    ch := buf[p];
  END;
  WHILE (decpt > 0) AND (i < size) DO
    s[i] := '0';
    INC( i );
    DEC( decpt );
  END;
  IF decpt > 0 THEN
    done := FALSE;
    RETURN;
  END;

  (* store the dec. point itself into result string *)
  s[i] := '.';
  INC( i );
  IF i >= size THEN
    done := FALSE;
    RETURN;
  END;

  (* now do the digits after the dec. point ... *)
  WHILE (decpt < 0) AND (digits > 0) AND (i < size) DO
    s[i] := '0';
    INC( i );
    INC( decpt );
    DEC( digits );
  END;
  ch := buf[p];
  WHILE (digits > 0) AND (i < size) AND (ch <> 0C) DO
    s[i] := ch;
    DEC( digits );
    INC( p );
    INC( i );
    ch := buf[p];
  END;
  WHILE (digits > 0) AND (i < size) DO
    s[i] := '0';
    DEC( digits );
    INC( i );
  END;
  IF digits > 0 THEN
    s[i] := 0C;
    done := FALSE;
    RETURN;
  END;

  (* finally insert leading blanks as needed *)
  s[i] := 0C;
  IF i < size THEN
    j := size - i;
    Fill( 0, j, ' ', s );
  END;
END FloatToFixPt;



PROCEDURE FloatToStr
(
  isLong   : BOOLEAN;
  r        : LONGREAL;
  digits   : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
VAR
  i        : LONGINT;
  p        : LONGINT;
  decpt    : INTEGER;
  sign     : BOOLEAN;
  buf      : DIGITSTRING;
  ch       : CHAR;
  highS    : LONGINT;
BEGIN
  highS := HIGH(s);
  IF digits >= 0 THEN
    (* Convert it into a fixed point format *)
    FloatToFixPt( isLong, r, SHORT(highS), digits, s, done );
    RETURN;
  ELSE
    digits := -digits;
  END;

  (* Convert it into a floating point format including exponent *)

  s[ highS ] := 0C;
  IF IsLongINFINITY(r) THEN
    (* it is an infinity *)
    COPY( "infinity", s );
    done := FALSE;
    RETURN;
  ELSIF IsLongNAN(r) THEN
    (* it is not a valid number *)
    COPY( "NaN", s );
    done := FALSE;
    RETURN;
  END;

  (* store significand digits into 'buf' *)
  floatcvt( isLong, FALSE, r, digits + 1, decpt, sign, buf, done );

  (* do the significand part: [-] digit . {digit} *)
  p := 0;
  i := 0;
  IF i >= highS THEN
    done := FALSE;
    RETURN;
  END;
  IF sign THEN
    s[i] := '-';
    INC( i );
  END;
  IF i >= highS THEN
    done := FALSE;
    RETURN;
  END;
  IF digits > 0 THEN
    s[i] := buf[p];
    INC( p );
  ELSE
    s[i] := '0';
  END;
  INC( i );
  IF i >= highS THEN
    done := FALSE;
    RETURN;
  END;
  s[i] := '.';
  INC( i );
  IF i >= highS THEN
    done := FALSE;
    RETURN;
  END;
  ch := buf[p];
  WHILE (digits > 0) AND (i < highS) AND (ch > 0C) DO
    s[i] := ch;
    INC( i );
    INC( p );
    ch := buf[p];
    DEC( digits );
  END;
  IF digits > 0 THEN
    done := FALSE;
    RETURN;
  END;

  (* now do the exponent part: E (+|-) digit digit digit *)
  IF r <> 0.0 THEN  (* avoid 0.00e-01 *)
    DEC( decpt );
  END;
  IF isLong AND (i+5 > highS) OR NOT isLong AND (i+4 > highS) THEN
    s[i] := 0C;
    done := FALSE;
    RETURN;
  END;
  s[i] := 'E';
  INC( i );
  IF decpt < 0 THEN
    s[i] := '-';
    decpt := -decpt;
  ELSE
    s[i] := '+';
  END;
  INC( i );
  IF isLong THEN
    s[i] := CHR( decpt DIV 100 + ORD('0') );
    INC( i );
    decpt := decpt MOD 100;
  END;
  s[i] := CHR( decpt DIV 10 + ORD('0') );
  INC( i );
  decpt := decpt MOD 10;
  s[i] := CHR( decpt + ORD('0') );
  INC( i );
  s[i] := 0C;
  RETURN;
END FloatToStr;



PROCEDURE RealToStr
( r        : REAL;
  digits   : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  (*$IF IsLongReal *)
  FloatToStr( IsLongReal, r, digits, s, done );
  (*$ELSE *)
  FloatToStr( IsLongReal, LONG(r), digits, s, done );
  (*$ENDIF *)
END RealToStr;



PROCEDURE LongRealToStr
( r        : LONGREAL;
  digits   : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  FloatToStr( TRUE, r, digits, s, done );
END LongRealToStr;



PROCEDURE ShortRealToStr
( r        : SHORTREAL;
  digits   : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  FloatToStr( FALSE, LONG( r ), digits, s, done );
END ShortRealToStr;



PROCEDURE RealToFixPt
( r        : REAL;
  size     : INTEGER;
  digits   : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  (*$IF IsLongReal *)
  FloatToFixPt( IsLongReal, r, size, digits, s, done );
  (*$ELSE *)
  FloatToFixPt( IsLongReal, LONG( r ), size, digits, s, done );
  (*$ENDIF *)
END RealToFixPt;



PROCEDURE LongRealToFixPt
( r        : LONGREAL;
  size     : INTEGER;
  digits   : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  FloatToFixPt( TRUE, r, size, digits, s, done );
END LongRealToFixPt;



PROCEDURE ShortRealToFixPt
( r        : SHORTREAL;
  size     : INTEGER;
  digits   : INTEGER;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  FloatToFixPt( FALSE, LONG( r ), size, digits, s, done );
END ShortRealToFixPt;



PROCEDURE StrToReal
( s        : ARRAY OF CHAR;
  VAR r    : REAL;
  VAR done : BOOLEAN
);
BEGIN
  (*
     Depending on the -R4 or -R8 compiler switch, the type size
     of REAL may be equal to that of SHORTREAL or LONGREAL.
     This procedure converts a string to real for both cases.
  *)
  (*$IF IsLongReal *)
  StrToLongReal( s, r, done );
  (*$ELSE *)
  StrToShortReal( s, r, done );
  (*$ENDIF *)
END StrToReal;



PROCEDURE StrToLongReal
( s        : ARRAY OF CHAR;
  VAR r    : LONGREAL;
  VAR done : BOOLEAN
);
VAR
  p       : LONGCARD;
  dval    : LONGREAL;
  exp     : INTEGER;
  msdec   : LONGCARD;
  lsdec   : LONGCARD;
  msscale : LONGCARD;
  dot     : BOOLEAN;
  sign    : BOOLEAN;
  subject : BOOLEAN;
  pow     : INTEGER;
  ch      : CHAR;
  sexp    : BOOLEAN;
  e       : INTEGER;
  u       : CARDINAL;
CONST
  MAXEXP  = 1024;
  HT      = CHR( 9  );
  LF      = CHR( 10 );
  VT      = CHR( 11 );
  FF      = CHR( 12 );
  CR      = CHR( 13 );
BEGIN
  IF Strings.Length( s ) > HIGH(s) THEN
    s[ HIGH(s) ] := 0C;
  END;
  p := 0;
  WHILE (s[p]=' ') OR (s[p]=HT) OR (s[p]=LF) OR (s[p]=VT) OR (s[p]=FF) OR (s[p]=CR) DO
    INC( p );
  END;
  sign := FALSE;              (* indicating + *)
  IF s[p] = '-' THEN
    sign := TRUE;
    INC( p );
  ELSIF s[p] = '+' THEN
    INC( p );
  END;
  subject := FALSE;
  dval := 0.0;
  dot := FALSE;               (* if decimal point has been seen *)
  exp := 0;
  msdec := 0;
  lsdec := 0;
  msscale := 1;
  LOOP
    ch := s[p];
    WHILE (('0' <= ch) AND (ch <= '9')) DO
      subject := TRUE; (* must have at least 1 digit *)
      IF msdec < (LONGCARD(MAX(LONGINT))-10) DIV 10 THEN
        msdec := msdec * 10 + VAL( LONGCARD, ORD( ch ) - ORD( '0' ) );
      ELSIF msscale < (LONGCARD(MAX(LONGINT))-10) DIV 10 THEN
        lsdec := lsdec * 10 + VAL( LONGCARD, ORD( ch ) - ORD( '0' ) );
        msscale := msscale * 10;
      ELSE
        INC( exp );
      END;
      IF dot THEN
        DEC( exp );
      END;
      INC( p );
      ch := s[p];
    END;
    IF (ch = '.') AND (NOT dot) THEN
      INC( p );
      dot := TRUE;
    ELSE
      EXIT;
    END;
  END;
  IF NOT subject THEN    (* if error (no digits seen) *)
    r := 0.0;            (* return 0.0 *)
    done := FALSE; RETURN;
  END;
  IF s[p] = 'E' THEN
    sexp := FALSE;
    INC( p );
    IF s[p] = '-' THEN
      sexp := TRUE;
      INC ( p );
    ELSIF s[p] = '+' THEN
      INC( p );
    END;
    subject := FALSE;
    e := 0;
    ch := s[p];
    WHILE ('0' <= ch) AND (ch <= '9') DO
      IF e < MAXEXP*2 THEN  (* prevent integer overflow *)
        e := e * 10 + VAL( INTEGER, ORD( ch ) - ORD( '0' ) );
      END;
      INC( p );
      subject := TRUE;
      ch := s[p];
    END;
    IF sexp THEN
      exp := exp - e;
    ELSE
      exp := exp + e;
    END;
    IF NOT subject THEN   (* if no digits in exponent *)
      r := 0.0;           (* return 0.0 *)
      done := FALSE; RETURN;
    END;
  END;
  dval := FLOAT( VAL( LONGINT, msdec ) );
  IF msscale <> 1 THEN   (* if stuff was accumulated in lsdec *)
    dval := dval * FLOAT( VAL( LONGINT, msscale ) );
    dval := dval + FLOAT( VAL( LONGINT, lsdec ) );
  END;
  IF dval <> 0.0 THEN
    u := 0;
    pow := 256;
    WHILE exp > 0 DO
      WHILE exp >= pow DO
        dval := dval * postab[ u ];
        exp := exp - pow;
      END;
      pow := pow DIV 2;
      INC( u );
    END;
    WHILE exp < 0 DO
      WHILE exp <= -pow DO
        dval := dval * negtab[ u ];
        IF dval = 0.0 THEN
          r := dval;
          done := FALSE; RETURN;
        END;
        exp := exp + pow;
      END;
      pow := pow DIV 2;
      INC( u );
    END;
  END;
  IF sign THEN
    r := -dval;
  ELSE
    r := dval;
  END;
  done := NOT IsLongNAN( dval ) AND NOT IsLongINFINITY( dval );
END StrToLongReal;



PROCEDURE StrToShortReal
( s        : ARRAY OF CHAR;
  VAR r    : SHORTREAL;
  VAR done : BOOLEAN
);
VAR
  dval    : LONGREAL;
  ok      : BOOLEAN;
BEGIN
  StrToLongReal( s, dval, ok );
  r := SHORT( dval );
  done := ok AND NOT IsShortNAN( r ) AND NOT IsShortINFINITY( r );
END StrToShortReal;



PROCEDURE RealToOctStr
(
  r        : REAL;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  (*$IF IsLongReal *)
  LongRealToOctStr( r, s, done );
  (*$ELSE *)
  ShortRealToOctStr( r, s, done );
  (*$ENDIF *)
END RealToOctStr;



PROCEDURE LongRealToOctStr
(
  r        : LONGREAL;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
CONST
  OctDigits = (TSIZE( LONGCARD ) * 8) DIV 3 + 1;
VAR
  hi        : LONGCARD;
  lo        : LONGCARD;
  s0        : ARRAY [0..OctDigits+1] OF CHAR;
  s1        : ARRAY [0..OctDigits+1] OF CHAR;
  len0      : LONGINT;
  len1      : LONGINT;
BEGIN
  s[0] := 0C;
  SYSTEM.INLINE
  (
    long i = java.lang.Double.doubleToLongBits(r);
    hi = (int)(i >>> 32);
    lo = (int)i;
  );
  Conversions.LongOctToStr( lo, s0, done );
  IF done THEN
    Conversions.LongOctToStr( hi, s1, done );
    IF done THEN
      len0 := Strings.Length( s0 );
      len1 := Strings.Length( s1 );
      IF s1[ len1-1 ] = 'B' THEN
        DEC( len1 );
        s1[ len1 ] := 0C;
        IF len1 < OctDigits THEN
          Strings.Fill( 0, OctDigits-len1, '0', s );
          len1 := OctDigits;
        END;
        Strings.Append( s1, s );
        IF len0 < OctDigits+1 THEN
          Strings.Fill( Strings.Length( s ), OctDigits+1-len0, '0', s );
          len0 := OctDigits+1;
        END;
        Strings.Append( s0, s );
        done := len0 + len1 = VAL( LONGINT, Strings.Length( s ) );
        RETURN;
      END;
    END;
  END;
  done := FALSE;
END LongRealToOctStr;



PROCEDURE ShortRealToOctStr
(
  r        : SHORTREAL;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
CONST
  OctDigits = (TSIZE( LONGCARD ) * 8) DIV 3 + 1;
VAR
  v         : LONGCARD;
  s0        : ARRAY [0..OctDigits+1] OF CHAR;
  len0      : LONGINT;
BEGIN
  s[0] := 0C;
  SYSTEM.INLINE
  (
    v = java.lang.Float.floatToIntBits(r);
  );
  Conversions.LongOctToStr( v, s0, done );
  IF done THEN
    len0 := Strings.Length( s0 );
    IF len0 < OctDigits+1 THEN
      Strings.Fill( 0, OctDigits+1-len0, '0', s );
    END;
    Strings.Append( s0, s );
    done := Strings.Length( s ) = OctDigits+1;
    RETURN;
  END;
  done := FALSE;
END ShortRealToOctStr;



PROCEDURE RealToHexStr
(
  r        : REAL;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
BEGIN
  (*$IF IsLongReal *)
  LongRealToHexStr( r, s, done );
  (*$ELSE *)
  ShortRealToHexStr( r, s, done );
  (*$ENDIF *)
END RealToHexStr;




PROCEDURE LongRealToHexStr
(
  r        : LONGREAL;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
CONST
  HexDigits = TSIZE( LONGCARD ) * 2;
VAR
  s0        : ARRAY [0..HexDigits+2] OF CHAR;
  s1        : ARRAY [0..HexDigits+2] OF CHAR;
  hi        : LONGCARD;
  lo        : LONGCARD;
  len1      : LONGINT;
  len0      : LONGINT;
BEGIN
  s[0] := 0C;
  SYSTEM.INLINE
  (
    long i = java.lang.Double.doubleToLongBits(r);
    hi = (int)(i >>> 32);
    lo = (int)i;
  );
  Conversions.LongHexToStr( lo, s0, done );
  IF done THEN
    Conversions.LongHexToStr( hi, s1, done );
    IF done THEN
      len1 := Strings.Length( s1 );
      len0 := Strings.Length( s0 );
      IF s1[ len1-1 ] = 'H' THEN
        DEC( len1 );
        s1[ len1 ] := 0C;
        IF len1 < HexDigits THEN
          Strings.Fill( 0, HexDigits-len1, '0', s );
          len1 := HexDigits;
        END;
        Strings.Append( s1, s );
        IF len0 < HexDigits+1 THEN
          Strings.Fill( Strings.Length( s ), HexDigits+1-len0, '0', s );
        ELSIF len0 > HexDigits+1 THEN
          Strings.Delete( s0, 0, len0-HexDigits-1 );
        END;
        len0 := HexDigits+1;
        Strings.Append( s0, s );
        done := len1 + len0 = VAL( LONGINT, Strings.Length( s ) );
        RETURN;
      END;
    END;
  END;
  done := FALSE;
END LongRealToHexStr;



PROCEDURE ShortRealToHexStr
(
  r        : SHORTREAL;
  VAR s    : ARRAY OF CHAR;
  VAR done : BOOLEAN
);
CONST
  HexDigits = TSIZE( LONGCARD ) * 2;
VAR
  v         : LONGCARD;
  s0        : ARRAY [0..HexDigits+2] OF CHAR;
  len0      : LONGINT;
BEGIN
  s[0] := 0C;
  SYSTEM.INLINE
  (
    v = java.lang.Float.floatToIntBits(r);
  );
  Conversions.LongHexToStr( v, s0, done );
  IF done THEN
    len0 := Strings.Length( s0 );
    IF len0 < HexDigits+1 THEN
      Strings.Fill( 0, HexDigits+1-len0, '0', s );
      len0 := HexDigits+1;
    END;
    Strings.Append( s0, s );
    done := VAL( LONGINT, Strings.Length( s ) ) = len0;
    RETURN;
  END;
  done := FALSE;
END ShortRealToHexStr;


END RealConversions.

