(*******************************************************************
   SYSTEM.MOD    Runtime kernel for Modula-2

   Copyright (c) 1998 by Mill Hill & Canterbury Corporation, Ltd.
********************************************************************)


IMPLEMENTATION MODULE SYSTEM;

(*$XL+ allow language extensions *)


VAR
  args : ADDRESS;


PROCEDURE OrBytes( Dest,Source1,Source2:ADDRESS; Size:LONGCARD );
VAR
  i,j       : LONGINT;
  val       : POINTER TO ARRAY OF SHORTCARD;
  val1      : POINTER TO ARRAY OF SHORTCARD;
  val2      : POINTER TO ARRAY OF SHORTCARD;
BEGIN
  (* Dest := Source1 OR Source2 for non-overlapping operands *)
  val := Dest;
  val1 := Source1;
  val2 := Source2;
  i := 0;
  j := Size;
  WHILE i < j DO
    val^[i] := val1^[i] OR val2^[i];
    INC( i );
  END;
END OrBytes;



PROCEDURE XorBytes( Dest,Source1,Source2:ADDRESS; Size:LONGCARD );
VAR
  i,j       : LONGINT;
  val       : POINTER TO ARRAY OF SHORTCARD;
  val1      : POINTER TO ARRAY OF SHORTCARD;
  val2      : POINTER TO ARRAY OF SHORTCARD;
BEGIN
  (* Dest := Source1 XOR Source2 for non-overlapping operands *)
  val := Dest;
  val1 := Source1;
  val2 := Source2;
  i := 0;
  j := Size;
  WHILE i < j DO
    val^[i] := val1^[i] XOR val2^[i];
    INC( i );
  END;
END XorBytes;


PROCEDURE AndBytes( Dest,Source1,Source2:ADDRESS; Size:LONGCARD );
VAR
  i,j       : LONGINT;
  val       : POINTER TO ARRAY OF SHORTCARD;
  val1      : POINTER TO ARRAY OF SHORTCARD;
  val2      : POINTER TO ARRAY OF SHORTCARD;
BEGIN
  (* Dest := Source1 AND Source2  for non-overlapping operands *)
  val := Dest;
  val1 := Source1;
  val2 := Source2;
  i := 0;
  j := Size;
  WHILE i < j DO
    val^[i] := val1^[i] AND val2^[i];
    INC( i );
  END;
END AndBytes;


PROCEDURE MemSet( Dest:ADDRESS; Val:BYTE; Size:LONGCARD );
VAR
  val : POINTER TO ARRAY OF SHORTCARD;
  i,j : LONGINT;
BEGIN
  (* Set multiple memory bytes to a given value *)
  i := 0;
  j := Size;
  val := Dest;
  WHILE i < j DO
    val^[i] := Val;
    INC( i );
  END;
END MemSet;


PROCEDURE MemCmp( Dest,Source: ADDRESS; Size:LONGCARD ) : LONGINT;
VAR
  val1 : POINTER TO ARRAY OF SHORTCARD;
  val2 : POINTER TO ARRAY OF SHORTCARD;
  i,j  : LONGINT;
BEGIN
  (* compare two memory regions and return result as follows:  *)
  (* -1: Dest < Source   0: Dest == Source   1: Dest > Source  *)
  val1 := Dest;
  val2 := Source;
  i := 0;
  j := Size;
  LOOP
    IF i >= j THEN
      EXIT;
    END;
    IF SHORTINT(val1^[i]) <> SHORTINT(val2^[i]) THEN
      EXIT;
    END;
    INC( i );
  END;
  IF i >= j THEN
    RETURN 0;
  ELSIF val1^[i] < val2^[i] THEN
    RETURN -1;
  ELSE
    RETURN 1;
  END;
END MemCmp;


PROCEDURE MemCpy( Dest,Source: ADDRESS; Size:LONGCARD );
VAR
  val1 : POINTER TO ARRAY OF SHORTCARD;
  val2 : POINTER TO ARRAY OF SHORTCARD;
  i,j  : LONGINT;
BEGIN
  (* copy muliple bytes from one memory region to another. *)
  val1 := Dest;
  val2 := Source;
  i := 0;
  j := Size;
  WHILE i < j DO
    val1^[i] := val2^[i];
    INC( i );
  END;
END MemCpy;


PROCEDURE SubsetOf( Set1, Set2 : ADDRESS; Size : LONGCARD ) : BOOLEAN;
BEGIN
  (* runtime function for implementing set1 <= set2 relation *)
  INLINE
  (
    byte[] t = new byte[Size];
    AndBytes( t, Set1, Set2, Size );
    return (MemCmp( t, Set1, Size ) == 0);
  )
END SubsetOf;


PROCEDURE AddSet( Set1, Set2 : ADDRESS; Size : LONGCARD ) : ADDRESS;
BEGIN
  (* runtime function for implementing set union *)
  INLINE
  (
    byte[] t = new byte[Size];
    OrBytes( t, Set1, Set2, Size );
    return t;
  );
END AddSet;


PROCEDURE SubSet( Set1, Set2 : ADDRESS; Size : LONGCARD ) : ADDRESS;
BEGIN
  (* runtime function for implementing set difference *)
  INLINE
  (
    byte[] t = new byte[Size];
    OrBytes( t, Set1, Set2, Size );
    XorBytes( t, t, Set2, Size );
    return t;
  );
END SubSet;



PROCEDURE MulSet( Set1, Set2 : ADDRESS; Size : LONGCARD ) : ADDRESS;
BEGIN
  (* runtime function for implementing set intersection *)
  INLINE
  (
    byte[] t = new byte[Size];
    AndBytes( t, Set1, Set2, Size );
    return t;
  );
END MulSet;


PROCEDURE DivSet( Set1, Set2 : ADDRESS; Size : LONGCARD ) : ADDRESS;
BEGIN
  (* runtime function for implementing set symmetric difference *)
  INLINE
  (
    byte[] t = new byte[Size];
    XorBytes( t, Set1, Set2, Size );
    return t;
  );
END DivSet;



PROCEDURE StrCmp( s1, s2 : ARRAY OF CHAR ) : LONGINT;
BEGIN
  (* runtime function for comparing 2 strings, returning: *)
  (* 1: s1 > s2    0: s1 == s2    -1: s1 < s2             *)
  INLINE
  (
    int j = s1.length;
    int k = s2.length;
    int i = 0;
    while ((i < j) && (i < k))
    {
      if (s1[i] == s2[i])
      {
        i++;
        continue;
      }
      if (s1[i] < s2[i])
      {
        return -1;
      }
      return 1;
    }
    if (j < k)
    {
      return -1;
    }
    if (j == k)
    {
      return 0;
    }
    return 1;
  );
END StrCmp;



PROCEDURE MUL_1( x,y : SHORTCARD ) : SHORTCARD;
BEGIN
  (* runtime function for unsigned 8-bit integer multiplication *)
  INLINE
  (
    int i,j,k;
    i = (int)x & 0x000000FF;
    j = (int)y & 0x000000FF;
    k = i * j;
    return (byte)k;
  );
END MUL_1;


PROCEDURE MUL_2( x,y : CARDINAL ) : CARDINAL;
BEGIN
  (* runtime function for unsigned 16-bit integer multiplication *)
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x0000FFFF;
    j = (int)y & 0x0000FFFF;
    k = i * j;
    return (short)k;
  );
END MUL_2;


PROCEDURE MUL_4( x,y : LONGCARD ) : LONGCARD;
BEGIN
  (* runtime function for unsigned 32-bit integer multiplication *)
  INLINE
  (
    long i,j,k;

    i = (long)x & 0xFFFFFFFFL;
    j = (long)y & 0xFFFFFFFFL;
    k = i * j;
    return (int)k;
  );
END MUL_4;


PROCEDURE DIV_1( x,y : SHORTCARD ) : SHORTCARD;
BEGIN
  (* runtime function for unsigned 8-bit integer division *)
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x000000FF;
    j = (int)y & 0x000000FF;
    k = i / j;
    return (byte)k;
  );
END DIV_1;


PROCEDURE DIV_2( x,y : CARDINAL ) : CARDINAL;
BEGIN
  (* runtime function for unsigned 16-bit integer division *)
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x0000FFFF;
    j = (int)y & 0x0000FFFF;
    k = i / j;
    return (short)k;
  );
END DIV_2;


PROCEDURE DIV_4( x,y : LONGCARD ) : LONGCARD;
BEGIN
  (* runtime function for unsigned 32-bit integer division *)
  INLINE
  (
    long i,j,k;

    i = (long)x & 0xFFFFFFFFL;
    j = (long)y & 0xFFFFFFFFL;
    k = i / j;
    return (int)k;
  );
END DIV_4;


PROCEDURE MOD_1( x,y : SHORTCARD ) : SHORTCARD;
BEGIN
  (* runtime function for unsigned 8-bit integer division remainder *)
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x000000FF;
    j = (int)y & 0x000000FF;
    k = i % j;
    return (byte)k;
  );
END MOD_1;


PROCEDURE MOD_2( x,y : CARDINAL ) : CARDINAL;
BEGIN
  (* runtime function for unsigned 16-bit integer division remainder *)
  INLINE
  (
    int i,j,k;

    i = (int)x & 0x0000FFFF;
    j = (int)y & 0x0000FFFF;
    k = i % j;
    return (short)k;
  );
END MOD_2;


PROCEDURE MOD_4( x,y : LONGCARD ) : LONGCARD;
BEGIN
  (* runtime function for unsigned 32-bit integer division remainder *)
  INLINE
  (
    long i,j,k;

    i = (long)x & 0xFFFFFFFFL;
    j = (long)y & 0xFFFFFFFFL;
    k = i % j;
    return (int)k;
  );
END MOD_4;


PROCEDURE CMP_1( x,y : SHORTCARD ) : LONGINT;
BEGIN
  (* runtime function for unsigned 8-bit integer compare *)
  INLINE
  (
    int i,j;

    i = (int)x & 0x000000FF;
    j = (int)y & 0x000000FF;
    if (i < j)
    {
      return -1;
    }
    if (i == j)
    {
      return 0;
    }
    return 1;
  );
END CMP_1;


PROCEDURE CMP_2( x,y : CARDINAL ) : LONGINT;
BEGIN
  (* runtime function for unsigned 16-bit integer compare *)
  INLINE
  (
    int i,j;

    i = (int)x & 0x0000FFFF;
    j = (int)y & 0x0000FFFF;
    if (i < j)
    {
      return -1;
    }
    if (i == j)
    {
      return 0;
    }
    return 1;
  );
END CMP_2;


PROCEDURE CMP_4( x,y : LONGCARD ) : LONGINT;
BEGIN
  (* runtime function for unsigned 32-bit integer compare *)
  INLINE
  (
    long i,j;

    i = (long)x & 0xFFFFFFFFL;
    j = (long)y & 0xFFFFFFFFL;
    if (i < j)
    {
      return -1;
    }
    if (i == j)
    {
      return 0;
    }
    return 1;
  );
END CMP_4;




PROCEDURE SetBitRange
(
  Dest    : ADDRESS;
  FromBit : LONGCARD;
  ToBit   : LONGCARD;
  MaxBit  : LONGCARD
);
VAR
  setval  : POINTER TO ARRAY OF SHORTCARD;
  i       : LONGCARD;
BEGIN
  setval := Dest;
  i := (HIGH( setval^ ) + 1) * 8 - 1;
  IF i < MaxBit THEN
    MaxBit := i;
  END;
  IF ToBit > MaxBit THEN
    ToBit := MaxBit;
  END;
  i := FromBit;
  WHILE i <= ToBit DO
    setval^[i DIV 8] := setval^[i DIV 8] OR VAL( SHORTCARD, (1 SHL (i MOD 8)) );
    i := i + 1;
  END;
END SetBitRange;



PROCEDURE DelBitRange
(
  Dest    : ADDRESS;
  FromBit : LONGCARD;
  ToBit   : LONGCARD;
  MaxBit  : LONGCARD
);
VAR
  setval  : POINTER TO ARRAY OF SHORTCARD;
  i       : LONGCARD;
BEGIN
  setval := Dest;
  i := (HIGH( setval^ ) + 1) * 8 - 1;
  IF i < MaxBit THEN
    MaxBit := i;
  END;
  IF ToBit > MaxBit THEN
    ToBit := MaxBit;
  END;
  i := FromBit;
  WHILE i <= ToBit DO
    setval^[i DIV 8] := setval^[i DIV 8] AND NOT VAL( SHORTCARD, (1 SHL (i MOD 8)) );
    i := i + 1;
  END;
END DelBitRange;



PROCEDURE TestBit( Set:ADDRESS; BitPos:LONGCARD; BitMax:LONGCARD ):BOOLEAN;
VAR
  setval  : POINTER TO ARRAY OF SHORTCARD;
  val     : SHORTCARD;
BEGIN
  (* Test a bit of a bitset; Return TRUE if set else return FALSE *)
  IF BitPos > BitMax THEN
    RETURN FALSE;
  END;
  setval := Set;
  val := setval^[ BitPos SHR 3 ];
  RETURN ( val AND (1 SHL (SHORTCARD( BitPos ) AND 7))) > 0;
END TestBit;




PROCEDURE Set1( s : PSET; i : LONGINT ) : PSET;
BEGIN
  (* runtime function for including a bit in a given set *)
  INLINE
  (
    s[i / 8] |= (byte)(1 << (i % 8));
    return s;
  );
END Set1;



PROCEDURE Set2( s : PSET; i : LONGINT; j : LONGINT ) : PSET;
BEGIN
  (* runtime function for including bits in a given set *)
  INLINE
  (
    int k = s.length * 8;
    while ((i <= j) && (i < k))
    {
      s[i / 8] |= (byte)(1 << (i % 8));
      i++;
    }
    return s;
  )
END Set2;



VAR
  ExitProc : PROC;
  NilProc  : PROC;


PROCEDURE GetExitProc() : PROC;
BEGIN
  RETURN ExitProc;
END GetExitProc;


PROCEDURE SetExitProc( UserProc : PROC );
BEGIN
  ExitProc := UserProc;
END SetExitProc;


PROCEDURE RunExitProc();
VAR
  SavedExitProc : PROC;
BEGIN
  SavedExitProc := NilProc;
  LOOP
    IF ExitProc = SavedExitProc THEN
      EXIT;
    END;
    IF ExitProc = NilProc THEN
      EXIT;
    END;
    SavedExitProc := ExitProc;
    ExitProc();
  END;
  ExitProc := NilProc;
END RunExitProc;



PROCEDURE ExitProgram( ExitCode : SHORTCARD );
BEGIN
  RunExitProc();
  INLINE
  (
    java.lang.System.runFinalization();
    java.lang.System.exit( ExitCode );
  );
END ExitProgram;



PROCEDURE CANCEL( e : ADDRESS );
BEGIN
  INLINE
  (
    ((java.lang.Exception)e).printStackTrace();
    if (args == null)
    {
      throw new java.lang.RuntimeException();
    }
    else
    {
      ExitProgram( (byte)1 );
    }
  );
END CANCEL;



PROCEDURE CAP( ch : CHAR ) : CHAR;
BEGIN
  INLINE
  (
    ch = java.lang.Character.toUpperCase(ch);
  );
  RETURN ch;
END CAP;



PROCEDURE IOTRANSFER( VAR p1 : ADDRESS; VAR p2 : ADDRESS; va : CARDINAL );
BEGIN
  (* not implemented for Java-Version of Modula-2 *)
END IOTRANSFER;



PROCEDURE LISTEN();
BEGIN
  (* not implemented for Java-Version of Modula-2 *)
END LISTEN;



PROCEDURE NEWPROCESS( p : PROC; a : ADDRESS; n : CARDINAL; VAR p1 : ADDRESS );
BEGIN
  (* not implemented for Java-Version of Modula-2 *)
END NEWPROCESS;



PROCEDURE TRANSFER( VAR p1 : ADDRESS; VAR p2 : ADDRESS );
BEGIN
  (* not implemented for Java-Version of Modula-2 *)
END TRANSFER;


PROCEDURE StartUp();
BEGIN
  (* put the start up stuff in here *)
  INLINE
  (
    NilProc = null;
  );
  ExitProc := NilProc;
END StartUp;


PROCEDURE TypeGuard( Wanted, Actual : ADDRESS );
BEGIN
  (* obsolete in Java-Version of Modula-2 *)
END TypeGuard;



PROCEDURE EnterPriority( Level : CARDINAL );
BEGIN
  (* This runtime function processes a priority-module entry.
     A priority module is to be protected against other coroutines,
     by giving it a higher priority level.
  *)
  INLINE
  (
    java.lang.Thread t = java.lang.Thread.currentThread();
    int i = Level & 0xFFFF;
    if (i > t.MAX_PRIORITY)
    {
      i = t.MAX_PRIORITY;
    }
    else if (i < t.MIN_PRIORITY)
    {
      i = t.MIN_PRIORITY;
    }
    t.setPriority( i );
  );
END EnterPriority;


PROCEDURE ExitPriority( Level : CARDINAL );
BEGIN
  (* This runtime function sets the current thread to default priority *)
  INLINE
  (
    java.lang.Thread t = java.lang.Thread.currentThread();
    int i = t.NORM_PRIORITY;
    t.setPriority( i );
  );
END ExitPriority;



PROCEDURE forName( name : ARRAY OF CHAR ) : ADDRESS;
VAR
  Result : ADDRESS;
BEGIN
  (* Runtime function for finding and loading a class *)
  (* Only used for passing the result to IsEqualProcs *)
  INLINE
  (
    try
    {
      Result = java.lang.Class.forName( new java.lang.String( name ) );
    }
    catch (java.lang.ClassNotFoundException e)
    {
      Result = null;
    }
  );
  RETURN Result;
END forName;



PROCEDURE IsEqualProcs( p1, p2 : ADDRESS ) : BOOLEAN;
VAR
  Result : BOOLEAN;
BEGIN
  (*
    Runtime function for testing equality of 2 procedure pointers:

      p1 = instance of procedure-pointer class or
           instance of procedure-pointer meta-class

      p2 = instance of procedure-pointer class or
           instance of procedure-pointer meta-class

    Procedure types are implemented as classes.
    A specific procedure address is implemented as an instance
    of an extended procedure type class. This extended class
    always has a name like this:

      <proc-type-id>_<mod-id>$ADR_<proc-id>

    Procedure pointer equality is tested by comparing the
    name parts starting with "$ADR".
  *)
  INLINE
  (
    java.lang.String s1;
    java.lang.String s2;
    int i1;
    int i2;

    while (true)
    {
      if ( (p1 instanceof java.lang.Class) &&
           (p2 instanceof java.lang.Class) ||
           (p1 instanceof java.lang.Object) &&
           (p2 instanceof java.lang.Object)
         )
      {
        if (p1 == p2)
        {
          Result = true;
          break;
        }
      }

      if (p1 instanceof java.lang.Class)
      {
        s1 = ((java.lang.Class)p1).getName();
      }
      else if (p1 instanceof java.lang.Object)
      {
        s1 = p1.getClass().getName();
      }
      else
      {
        Result = (p1 == p2);
        break;
      }

      if (p2 instanceof java.lang.Class)
      {
        s2 = ((java.lang.Class)p2).getName();
      }
      else if (p2 instanceof java.lang.Object)
      {
        s2 = p2.getClass().getName();
      }
      else
      {
        Result = (p1 ==p2);
        break;
      }

      if (s1.equals( s2 ))
      {
        Result = true;
        break;
      }

      i1 = s1.indexOf( "_ADR" );
      if (i1 < 0)
      {
        Result = false;
        break;
      }
      i2 = s2.indexOf( "_ADR" );
      if (i2 < 0)
      {
        Result = false;
        break;
      }
      s1 = s1.substring( i1 );
      s2 = s2.substring( i2 );
      Result = (s1.equals( s2 ));
      break;
    }
  );
  RETURN Result;
END IsEqualProcs;


PROCEDURE InitArray( arrayref : ADDRESS );
BEGIN
  (* Runtime function for allocating structured array components *)
  INLINE
  (
    int i,j;
    java.lang.Class ComponentType = arrayref.getClass().getComponentType();
    boolean is_array_component = ComponentType.isArray();
    java.lang.Object[] arrayobj = (java.lang.Object[])arrayref;
    j = arrayobj.length;
    for (i=0; i<j; i++)
    {
      if (is_array_component)
      {
        InitArray( arrayobj[i] );
      }
      else
      {
        try
        {
          arrayobj[i] = ComponentType.newInstance();
        }
        catch (java.lang.InstantiationException e)
        {
          throw new java.lang.RuntimeException
          ( "Array component not created because of InstantiationException" );
        }
        catch (java.lang.IllegalAccessException e)
        {
          throw new java.lang.RuntimeException
          ( "Array component not created because of IllegalAccessException" );
        }
      }
    }
  );
END InitArray;



PROCEDURE CloneArray( arrayref : ADDRESS ) : ADDRESS;
VAR
  Result : ADDRESS;
BEGIN
  (* Runtime function for non-shallow array cloning,
     only used for compiler versions >= 3.0
  *)
  INLINE
  (

    int                      i,j;
    java.lang.Class[]        signature;
    java.lang.reflect.Method clone_method = null;
    java.lang.Class          ComponentType = arrayref.getClass().getComponentType();
    java.lang.Object[]       arrayobj;

    if (ComponentType.isArray())
    {
      arrayref = ((java.lang.Object[])arrayref).clone();
      arrayobj = (java.lang.Object[])arrayref;
      j = arrayobj.length;
      for (i=0; i<j; i++)
      {
        arrayobj[i] = CloneArray( arrayobj[i] );
      }
      Result = arrayobj;
    }
    else if (ComponentType == java.lang.Boolean.TYPE)
    {
      Result = ((boolean[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Character.TYPE)
    {
      Result = ((char[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Byte.TYPE)
    {
      Result = ((byte[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Short.TYPE)
    {
      Result = ((short[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Integer.TYPE)
    {
      Result = ((int[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Long.TYPE)
    {
      Result = ((long[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Float.TYPE)
    {
      Result = ((float[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Double.TYPE)
    {
      Result = ((double[])arrayref).clone();
    }
    else if (ComponentType == java.lang.Void.TYPE)
    {
      Result = arrayref;
      SYSTEM.CANCEL
      (
        new java.lang.RuntimeException( "Invalid array of void" )
      );
    }
    else /* ComponentType is a class or interface */
    {
      signature = new java.lang.Class[1];
      signature[0] = java.lang.Void.TYPE;
      arrayref = ((java.lang.Object[])arrayref).clone();
      arrayobj = (java.lang.Object[])arrayref;
      try
      {
        clone_method = ComponentType.getMethod( "_Clone", signature );
        if (clone_method == null)
        {
          clone_method = ComponentType.getMethod( "Clone", signature );
        }
        if (clone_method != null)
        {
          j = arrayobj.length;
          for (i=0; i<j; i++)
          {
            arrayobj[i] = clone_method.invoke( arrayobj[i], null );
          }
        }
      }
      catch (java.lang.Exception e)
      {
        SYSTEM.CANCEL( e );
      }
      if (clone_method == null)
      {
        throw new java.lang.RuntimeException
        ( "Array component(s) are not cloneable" );
      }
      Result = arrayobj;
    }
  );
  RETURN Result;
END CloneArray;



PROCEDURE CopyArray( dest : ADDRESS; origin : ADDRESS );
BEGIN
  (* Runtime procedure for non-shallow array copy,
     only used for compiler versions >= 3.0
  *)
  INLINE
  (
    int                      i,j;
    java.lang.Class          ComponentType = dest.getClass().getComponentType();
    java.lang.Object[]       to;
    java.lang.Object[]       from;
    java.lang.Class[]        signature;
    java.lang.reflect.Method copy_method = null;
    java.lang.Object[]       args;

    if (ComponentType.isArray())
    {
      to = ((java.lang.Object[])dest);
      from = ((java.lang.Object[])origin);
      j = to.length;
      for (i=0; i<j; i++)
      {
        CopyArray( to[i], from[i] );
      }
    }
    else if (ComponentType == java.lang.Boolean.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((boolean[])dest).length );
    }
    else if (ComponentType == java.lang.Character.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((char[])dest).length );
    }
    else if (ComponentType == java.lang.Byte.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((byte[])dest).length );
    }
    else if (ComponentType == java.lang.Short.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((short[])dest).length );
    }
    else if (ComponentType == java.lang.Integer.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((int[])dest).length );
    }
    else if (ComponentType == java.lang.Long.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((long[])dest).length );
    }
    else if (ComponentType == java.lang.Float.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((float[])dest).length );
    }
    else if (ComponentType == java.lang.Double.TYPE)
    {
      java.lang.System.arraycopy
      ( origin, 0, dest, 0, ((double[])dest).length );
    }
    else if (ComponentType == java.lang.Void.TYPE)
    {
      SYSTEM.CANCEL
      (
        new java.lang.RuntimeException( "Invalid array of void" )
      );
    }
    else /* ComponentType is a class or interface */
    {
      signature = new java.lang.Class[1];
      signature[0] = ComponentType;
      to = (java.lang.Object[])dest;
      from = (java.lang.Object[])origin;
      java.lang.System.arraycopy( from, 0, to, 0, to.length );
      try
      {
        copy_method = ComponentType.getMethod( "_Copy", signature );
        if (copy_method != null)
        {
          j = to.length;
          for (i=0; i<j; i++)
          {
            args = new java.lang.Object[1];
            args[0] = from[i];
            copy_method.invoke( to[i], args );
          }
        }
      }
      catch (java.lang.Exception e)
      {
        SYSTEM.CANCEL( e );
      }
      if (copy_method == null)
      {
        throw new java.lang.RuntimeException
        ( "Array component(s) have no method _Copy" );
      }
    }
  );
END CopyArray;



BEGIN (* SYSTEM *)
  StartUp();
  INLINE
  (
    args = _args;
  );
END SYSTEM.

